#include <cstdio>
#include <iostream>
#include <fstream>
#include <vector>
#include "hashgrid.h"
#include "sparsearray3.h"
#include "vec3f.h"

using namespace std;

struct QuadMesh
{
   vector<Vec3f> vertex, normal;
   vector<IntTuple<4> > face;
};


bool read_frame(const char *filename, vector<Vec3f> &x, float &radius)
{
   ifstream in(filename);
   if(!in.good())
      return false;
   float junk; in>>junk;
   in>>radius;
   int n;
   in>>n;
   x.resize(n);
   for(int i=0; i<n; ++i)
      in>>x[i];
   return true;
}

void get_edge(QuadMesh &mesh, HashTable<IntTuple<6>, int> &intersections, float dx,
              float F0, int i0, int j0, int k0, float F1, int i1, int j1, int k1, int &vertex_index)
{
   IntTuple<6> key(i0, j0, k0, i1, j1, k1);
   if(i0>i1 || (i0==i1 && j0>j1) || (i0==i1 && j0==j1 && k0>k1)){
      key.i[0]=i1; key.i[1]=j1; key.i[2]=k1;
      key.i[3]=i0; key.i[4]=j0; key.i[5]=k0;
   }
   if(!intersections.get_entry(key, vertex_index)){
      vertex_index=mesh.vertex.size();
      intersections.add(key, vertex_index);
      mesh.vertex.push_back(dx*Vec3f(i0-F0*(i1-i0)/(F1-F0), j0-F0*(j1-j0)/(F1-F0), k0-F0*(k1-k0)/(F1-F0)));
   }
}

void check_tet(QuadMesh &mesh, const SparseArray3 &f, HashTable<IntTuple<6>, int> &intersections, float threshhold, float dx,
               int i0, int j0, int k0, int i1, int j1, int k1, 
               int i2, int j2, int k2, int i3, int j3, int k3)
{
   float F0=threshhold-f(i0,j0,k0), F1=threshhold-f(i1,j1,k1), F2=threshhold-f(i2,j2,k2), F3=threshhold-f(i3,j3,k3);
   int v01, v02, v03, v12, v13, v23;

   // look for the sign changes
   if(F3<0){
      if(F2<0){
         if(F1<0){
            if(F0<0){
               ;//nothing
            }else{ // F0>=0
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F1,i1,j1,k1, v01);
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F2,i2,j2,k2, v02);
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F3,i3,j3,k3, v03);
               mesh.face.push_back(IntTuple<4>(v01, v02, v03, -1));
            }
         }else{ // F1>=0
            if(F0<0){
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F1,i1,j1,k1, v01);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F2,i2,j2,k2, v12);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F3,i3,j3,k3, v13);
               mesh.face.push_back(IntTuple<4>(v01, v13, v12, -1));
            }else{ // F0>=0
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F2,i2,j2,k2, v02);
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F3,i3,j3,k3, v03);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F2,i2,j2,k2, v12);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F3,i3,j3,k3, v13);
               mesh.face.push_back(IntTuple<4>(v02, v12, v13, v03));
            }
         }
      }else{ // F2>=0
         if(F1<0){
            if(F0<0){
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F2,i2,j2,k2, v02);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F2,i2,j2,k2, v12);
               get_edge(mesh, intersections, dx, F2,i2,j2,k2, F3,i3,j3,k3, v23);
               mesh.face.push_back(IntTuple<4>(v02, v12, v23, -1));
            }else{ // F0>=0
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F1,i1,j1,k1, v01);
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F3,i3,j3,k3, v03);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F2,i2,j2,k2, v12);
               get_edge(mesh, intersections, dx, F2,i2,j2,k2, F3,i3,j3,k3, v23);
               mesh.face.push_back(IntTuple<4>(v01, v03, v23, v12));
            }
         }else{ // F1>=0
            if(F0<0){
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F1,i1,j1,k1, v01);
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F2,i2,j2,k2, v02);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F3,i3,j3,k3, v13);
               get_edge(mesh, intersections, dx, F2,i2,j2,k2, F3,i3,j3,k3, v23);
               mesh.face.push_back(IntTuple<4>(v01, v02, v23, v13));
            }else{ // F0>=0
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F3,i3,j3,k3, v03);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F3,i3,j3,k3, v13);
               get_edge(mesh, intersections, dx, F2,i2,j2,k2, F3,i3,j3,k3, v23);
               mesh.face.push_back(IntTuple<4>(v03, v13, v23, -1));
            }
         }
      }
   }else{ // F3>=0
      if(F2<0){
         if(F1<0){
            if(F0<0){
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F3,i3,j3,k3, v03);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F3,i3,j3,k3, v13);
               get_edge(mesh, intersections, dx, F2,i2,j2,k2, F3,i3,j3,k3, v23);
               mesh.face.push_back(IntTuple<4>(v03, v23, v13, -1));
            }else{ // F0>=0
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F1,i1,j1,k1, v01);
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F2,i2,j2,k2, v02);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F3,i3,j3,k3, v13);
               get_edge(mesh, intersections, dx, F2,i2,j2,k2, F3,i3,j3,k3, v23);
               mesh.face.push_back(IntTuple<4>(v01, v13, v23, v02));
            }
         }else{ // F1>=0
            if(F0<0){
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F1,i1,j1,k1, v01);
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F3,i3,j3,k3, v03);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F2,i2,j2,k2, v12);
               get_edge(mesh, intersections, dx, F2,i2,j2,k2, F3,i3,j3,k3, v23);
               mesh.face.push_back(IntTuple<4>(v01, v12, v23, v03));
            }else{ // F0>=0
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F2,i2,j2,k2, v02);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F2,i2,j2,k2, v12);
               get_edge(mesh, intersections, dx, F2,i2,j2,k2, F3,i3,j3,k3, v23);
               mesh.face.push_back(IntTuple<4>(v02, v23, v12, -1));
            }
         }
      }else{ // F2>=0
         if(F1<0){
            if(F0<0){
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F2,i2,j2,k2, v02);
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F3,i3,j3,k3, v03);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F2,i2,j2,k2, v12);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F3,i3,j3,k3, v13);
               mesh.face.push_back(IntTuple<4>(v02, v03, v13, v12));
            }else{ // F0>=0
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F1,i1,j1,k1, v01);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F2,i2,j2,k2, v12);
               get_edge(mesh, intersections, dx, F1,i1,j1,k1, F3,i3,j3,k3, v13);
               mesh.face.push_back(IntTuple<4>(v01, v12, v13, -1));
            }
         }else{ // F1>=0
            if(F0<0){
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F1,i1,j1,k1, v01);
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F2,i2,j2,k2, v02);
               get_edge(mesh, intersections, dx, F0,i0,j0,k0, F3,i3,j3,k3, v03);
               mesh.face.push_back(IntTuple<4>(v01, v03, v02, -1));
            }else{ // F0>=0
               ;//nothing
            }
         }
      }
   }
}

void compute_mesh_for_frame(const vector<Vec3f> &x, float radius, float dx, QuadMesh &mesh, float threshhold)
{
   cout<<"  sampling implicit function..."<<endl;
   cout<<"    "; cout.flush();
   SparseArray3 f; // the implicit function we will construct
   int milestone=x.size()/60;
   for(unsigned int p=0; p<x.size(); ++p){ // loop over particles
      // the list of grid cells this particle may influence
      int ilow=(int)((x[p][0]-radius)/dx)-1, ihigh=(int)((x[p][0]+radius)/dx)+1;
      int jlow=(int)((x[p][1]-radius)/dx)-1, jhigh=(int)((x[p][1]+radius)/dx)+1;
      int klow=(int)((x[p][2]-radius)/dx)-1, khigh=(int)((x[p][2]+radius)/dx)+1;
      // loop over them, adding to f (including some zeros - we'll need them for later)
      for(int i=ilow; i<=ihigh; ++i) for(int j=jlow; j<=jhigh; ++j) for(int k=klow; k<=khigh; ++k){
         float d=mag(x[p]-dx*Vec3f(i, j, k))/radius;
         //f(i,j,k)+=(d<1) ? (2*d-3)*d*d+1 : 0;
	 if(d<1){
            f(i,j,k)+=(2*d-3)*d*d+1;
	 }
      }
      if(p%milestone==0) cout<<"."; cout.flush();
   }
   cout<<endl;
   f.A.output_statistics();
   cout<<"------------------------------"<<endl;

   // now do marching tets on the grid
   cout<<"  doing marching tetrahedra..."<<endl;
   cout<<"    "; cout.flush();
   HashTable<IntTuple<6>, int> intersections;
   milestone=f.A.size()/60;
   int p=0;
   for(HashTable<IntTuple<3>, float>::iterator c=f.A.begin(); c!=f.A.end(); ++c){
      int i=c->key.i[0], j=c->key.i[1], k=c->key.i[2];
      if((i+j+k)%2 == 0){
         check_tet(mesh, f, intersections, threshhold, dx, i,j,k, i+1,j,k, i,j,k+1, i,j+1,k);
         check_tet(mesh, f, intersections, threshhold, dx, i+1,j,k, i+1,j,k+1, i,j,k+1, i+1,j+1,k+1);
         check_tet(mesh, f, intersections, threshhold, dx, i+1,j+1,k, i+1,j+1,k+1, i+1,j,k, i,j+1,k);
         check_tet(mesh, f, intersections, threshhold, dx, i,j+1,k, i+1,j+1,k+1, i,j,k+1, i,j+1,k+1);
         check_tet(mesh, f, intersections, threshhold, dx, i,j+1,k, i+1,j+1,k+1, i+1,j,k, i,j,k+1);
      }else{
         check_tet(mesh, f, intersections, threshhold, dx, i,j,k, i+1,j+1,k, i,j+1,k+1, i,j+1,k);
         check_tet(mesh, f, intersections, threshhold, dx, i,j,k, i+1,j,k, i+1,j,k+1, i+1,j+1,k);
         check_tet(mesh, f, intersections, threshhold, dx, i,j,k, i,j+1,k+1, i+1,j,k+1, i,j,k+1);
         check_tet(mesh, f, intersections, threshhold, dx, i,j,k, i+1,j+1,k, i+1,j,k+1, i,j+1,k+1);
         check_tet(mesh, f, intersections, threshhold, dx, i+1,j+1,k+1, i,j+1,k+1, i+1,j,k+1, i+1,j+1,k);
      }
      if(p%milestone==0) cout<<"."; cout.flush();
      ++p;
   }
   cout<<endl;
   cout<<"---- intersection table at end"<<endl;
   intersections.output_statistics();
   cout<<"------------------------------"<<endl;
}

void compute_normals_for_frame(const vector<Vec3f> &x, float radius, QuadMesh &mesh)
{
   cout<<"    "; cout.flush();
   mesh.normal.resize(mesh.vertex.size());

   float nradius=1.25*radius;
   HashGrid3D grid(x, nradius);
   grid.update();
   int j;
   int milestone=mesh.vertex.size()/60;
   for(unsigned int i=0; i<mesh.vertex.size(); ++i){ // get the normal at each mesh vertex
      Vec3f meshx(mesh.vertex[i]);
      mesh.normal[i]=Vec3f(0,0,0);
      for(HashGrid3D::finder p(grid, meshx, nradius); p.still_going(); ++p){
	 j=*p;
	 Vec3f dx(x[j]-meshx);
         float d=mag(dx)/nradius;
	 if(d>0 && d<1)
	    mesh.normal[i]+=(6*d*(d-1)/(mag(dx)*nradius)) * dx;
      }
      float m=mag(mesh.normal[i]);
      if(m==0){
	 cout<<"ZERO NORMAL"<<endl;
	 mesh.normal[i]=Vec3f(1,0,0);
      }else
	 mesh.normal[i]*=1.0/m;
      if(i%milestone==0) cout<<"."; cout.flush();
   }
   cout<<endl;
   cout<<"---- hashgrid statistics"<<endl;
   grid.grid.output_statistics();
   cout<<"------------------------------"<<endl;
}

void read_camera(const char *camfile, char *&cam)
{
   ifstream in(camfile);
   vector<char> camvec;
   for(;;){
      char c=in.get();
      if(in.good()) camvec.push_back(c);
      else break;
   }
   cam=new char[camvec.size()+1];
   unsigned int i;
   for(i=0; i<camvec.size(); ++i)
      cam[i]=camvec[i];
   cam[i]=0;
}

void output_frame(const char *animation_name, const char *camera, const QuadMesh &mesh)
{
   // figure out the RIB file name, and open it
   char *filename=new char[strlen(animation_name)+20];
   sprintf(filename, "%s.rib", animation_name);
   ofstream out(filename);

   // first line: what image file this RIB file should produce
   sprintf(filename, "%s.tiff", animation_name);
   out<<"Display \""<<filename<<"\" \"file\" \"rgb\""<<endl;
   delete[] filename;

   // next line: image format (width and height in pixels, pixel aspect ratio)
   out<<"Format 640 480 1"<<endl;

   // sampling rate - 2x2 isn't too bad
   out<<"PixelSamples 2 2"<<endl;

   // correct for a monitor with gamma 2.2 (we won't be compositing these images!)
   out<<"Exposure 1.0 2.2"<<endl;

   // then write out the camera specification
   out<<camera<<endl;

   // start the scene
   out<<"WorldBegin"<<endl;

   // lights
   out<<"LightSource \"ambientlight\" 1 \"intensity\" .1 \"lightcolor\" [1 1 1]"<<endl;
   out<<"LightSource \"distantlight\" 2 \"intensity\" 0.8 \"from\" [-5 30 20] \"to\" [0 0 0]"<<endl;
   out<<"LightSource \"distantlight\" 2 \"intensity\" 0.3 \"from\" [-5 -10 20] \"to\" [0 0 0]"<<endl;

   // set up material for mud
   out<<"Color [1 0.7 0.2]"<<endl;
   out<<"Surface \"plastic\" \"Ka\" 0.1 \"specularcolor\" [1 1 1]"<<endl;

   // for debugging, and speed!
   //out<<"Sides 1"<<endl;

   // output the geometry
   unsigned int i;
   out<<"PointsPolygons ["<<endl; // first the number of vertices in each face
   for(i=0; i<mesh.face.size(); ++i){
      if(mesh.face[i].i[3]==-1)
	 out<<" 3";
      else
	 out<<" 4";
      if(i%35==34) out<<endl;
   }
   out<<"] ["<<endl; // now the vertex index list for each face
   for(i=0; i<mesh.face.size(); ++i){
      out<<" "<<mesh.face[i].i[0]<<" "<<mesh.face[i].i[1]<<" "<<mesh.face[i].i[2];
      if(mesh.face[i].i[3]>=0)
         out<<" "<<mesh.face[i].i[3];
      out<<endl;
   }
   out<<"] \"P\" ["<<endl; // and now the vertex coordinates
   for(i=0; i<mesh.vertex.size(); ++i){
      out<<" "<<mesh.vertex[i]<<endl;
   }
   out<<"] \"N\" ["<<endl; // and then the normal coordinates
   for(i=0; i<mesh.normal.size(); ++i){
      out<<" "<<mesh.normal[i]<<endl;
   }
   out<<"]"<<endl; // and then the normal coordinates


   // finish the scene
   out<<"WorldEnd"<<endl;
}

int main(int argc, char **argv)
{
   if(argc<3){
      cerr<<"Expecting two arguments at least: <animation_filename> <camera_filename> [sampling_rate,default=0.3]"<<endl;
      return 1;
   }

   vector<Vec3f> x;
   float radius;
   cout<<"reading animation file..."<<endl;
   if(!read_frame(argv[1], x, radius)){
      cerr<<"couldn't read animation file "<<argv[1]<<endl;
      return 1;
   }

   float sampling_rate=0.3;
   if(argc==4){
      sscanf(argv[3], "%f", &sampling_rate);
   }
   cout<<"computing mesh..."<<endl;
   QuadMesh mesh;
   compute_mesh_for_frame(x, radius, sampling_rate*radius, mesh, 0.5);

   cout<<"computing normals..."<<endl;
   compute_normals_for_frame(x, radius, mesh);

   cout<<"reading camera..."<<endl;
   char *camera=0;
   read_camera(argv[2], camera);

   cout<<"writing RIB file..."<<endl;
   output_frame(argv[1], camera, mesh);
   return 0;
}

